let dp_init_cache = ref 128;;

let dp_max_cache = ref 65536;;

type bdd_ns = {
  node_table: (int * int * int) Vector.t;
  inverse: ((int * int * int),  int) Hashtbl.t;
};;

let make_ns sz_hint =
  let tt = Vector.create ~sz:sz_hint (-1,-1,-1) 
  and ht = Hashtbl.create sz_hint 
  in begin
    Vector.append tt (sz_hint + 1, -1, -1);
    Vector.append tt (sz_hint, -1, -1);
    { node_table=tt; inverse=ht; }
  end;;

let make ns var low high =
  if low = high then 
    low
  else 
    if Hashtbl.mem ns.inverse (var, low, high) then
      Hashtbl.find ns.inverse (var, low, high)
    else
      let max_var, _, _ = Vector.get ns.node_table 0 in
      let _ = Vector.append ns.node_table (var, low, high) in
      let node = (Vector.length ns.node_table) - 1 in
      let _ = Hashtbl.add ns.inverse (var, low, high) node 
      in  node;;

let bddmax ns = (Vector.length ns.node_table)-1;;

let ap_and = function
    (0,_) -> 0
  | (1,y) -> y
  | (x,0) -> 0
  | (x,1) -> x
  | _     -> failwith "impossible";;

let ap_or = function
    (0,y) -> y
  | (1,y) -> 1
  | (x,0) -> x
  | (x,1) -> 1
  | _     -> failwith "impossible";;

let apply op ns a b =
  let ht = Hashtbl.create !dp_init_cache in
  let rec app n1 n2 =
    if Hashtbl.mem ht (n1, n2) then
      Hashtbl.find ht (n1, n2)
    else if (n1 < 2 || n2 < 2) then
      op (n1,n2)
    else 
      begin
      let vn1, lown1, highn1 = Vector.get ns.node_table n1
      and vn2, lown2, highn2 = Vector.get ns.node_table n2 in
      let next = 
	if vn1 = vn2 then
	  make ns vn1 (app lown1 lown2) (app highn1 highn2)
	else if vn1 < vn2 then
	  make ns vn1 (app lown1 n2) (app highn1 n2)
	else
	  make ns vn2 (app n1 lown2) (app n1 highn2)
      in  Hashtbl.add ht (n1, n2) next;
	  next
      end
  in (app a b);;

let bddand ns x a = apply ap_and ns x a;;
let bddor ns x a = apply ap_or ns x a;;

let bddadd ns x a b = 
  let c = apply ap_and ns a b 
  in  apply ap_or ns x c;;

let bddvar  ns v = make ns v 0 1;;
let bddnvar ns v = make ns v 1 0;;
let bddtop ns = 1;;
let bddbot ns = 0;;

module ISet = Set.Make(
  struct 
    type t = int 
    let compare x y = -(compare x y) 
  end);;

let rec bddnodes ns n =
  if n<2 then ISet.empty
  else let _, low, high = Vector.get ns.node_table n in
       let ls = bddnodes ns low
       and hs = bddnodes ns high 
       in  ISet.union (ISet.union ls hs) (ISet.singleton n);;

let bddaff ns (n:int) = 
  let nodes = ISet.elements (bddnodes ns n)
  and f x = let v, low, high = Vector.get ns.node_table x
            in (Printf.printf "node %2d var %2d low %2d high %2d\n" x v low high)
  in  begin
        Printf.printf "BDD start %d\n------------------\n" n;
        List.iter f nodes;
        Printf.printf "------------------\n"
      end;;

(* let bddcopy ns upto n =
  let nodes = bddnodes ns n 
  and ht    = Hashtbl.create !dp_init_cache in
  let nns   = make_ns ((ISet.cardinal nodes)+2)
  and l     = List.rev (ISet.elements nodes) in
  let add x =     
    if Hashtbl.mem ht x then ()
    else 
      let v, low, high = Vector.get ns.node_table x in  
      let a = Hashtbl.find ht low
      and b = Hashtbl.find ht high in
      let y = make nns v a b
      in  Hashtbl.add ht x y 
  in begin
    Hashtbl.add ht 0 0;
    Hashtbl.add ht 1 1;
    for i = 0 to upto do
      add i
    done;
    List.iter add l;
    (nns,Hashtbl.find ht n)
  end;; *)

let bddcopy ns upto n =
  let ht  = Hashtbl.create !dp_init_cache in
  let nns = make_ns (upto+2) in
  let rec loop x =
    if Hashtbl.mem ht x 
    then
      Hashtbl.find ht x
    else 
      let v, low, high = Vector.get ns.node_table x in  
      let a = loop low  in
      let b = loop high in
      let y = make nns v a b
      in  (Hashtbl.add ht x y; y) in
  let add x = let _ = loop x in ()
  in begin
    Hashtbl.add ht 0 0;
    Hashtbl.add ht 1 1;
    for i = 2 to upto do
      add i
    done;
    add n;
    (nns,Hashtbl.find ht n)
  end;;

let bddistop _ x   = (x=1);;
let bddisbot _ x   = (x=0);;
let bddeq    _ x y = (x=y);;
let bddneq   _ x y = (x!=y);;

let bddcmodel ns a =
  let rec amod node assgn =
    if node = 1 then None
    else if node = 0 then 
       Some assgn
    else 
      let vn, ln, hn = Vector.get ns.node_table node 
      in if ln = 1 
         then amod hn ((vn,true)::assgn)
	 else amod ln ((vn,false)::assgn)
  in amod a [];;

let bddmodel ns a =
  let rec amod node assgn =
    if node = 0 then None
    else if node = 1 then 
       Some assgn
    else 
      let vn, ln, hn = Vector.get ns.node_table node 
      in if ln = 0 
         then amod hn ((vn,true)::assgn)
	 else amod ln ((vn,false)::assgn)
  in amod a [];;

class type bdd_t =
  object
    method max     : int
    method conj    : int -> int -> int
    method disj    : int -> int -> int
    method add     : int -> int -> int -> int
    method var     : int -> int
    method nvar    : int -> int
    method top     : int
    method bot     : int
    method istop   : int -> bool
    method isbot   : int -> bool
    method eq      : int -> int -> bool
    method neq     : int -> int -> bool
    method cmodel  : int -> (int*bool) list option
    method model   : int -> (int*bool) list option
    method print   : int -> unit
    method table   : int -> (int*int*int)
    method copy    : bdd_t
    method import  : bdd_t -> int -> int
    method extract : int -> int -> (bdd_t*int)
  end;;

let bddimport nns (ons:bdd_t) n =
  let ht  = Hashtbl.create !dp_init_cache in
  let _   = Hashtbl.add ht 0 0            in
  let _   = Hashtbl.add ht 1 1            in
  let rec loop x =
     try Hashtbl.find ht x
     with Not_found -> 
         let v,l,h = ons#table x in
         let a = loop l          in
         let b = loop h          in
         let y = make nns v a b  in
         let _ = Hashtbl.add ht x y
         in  y
  in loop n;;

class copybdd ns =
  object (s:#bdd_t)
     method max     = bddmax ns
     method conj    = bddand ns
     method disj    = bddor ns
     method add     = bddadd ns
     method var     = bddvar ns
     method nvar    = bddnvar ns
     method top     = bddtop ns
     method bot     = bddbot ns
     method istop   = bddistop ns
     method isbot   = bddisbot ns
     method eq      = bddeq ns
     method neq     = bddneq ns
     method cmodel  = bddcmodel ns
     method model   = bddmodel ns
     method print   = bddaff ns
     method table   = Vector.get ns.node_table
     method import ons n = bddimport ns ons n
     method copy    =
         let nns = { node_table = Vector.copy ns.node_table;
                     inverse    = Hashtbl.copy ns.inverse;
                   }
         in  new copybdd nns
     method extract u x = 
         let (nns,y) = bddcopy ns u x 
         in (new copybdd nns,y)  
  end;;

class emptybdd n = let ns = make_ns n in copybdd ns;;

