#load "libs.cma";;

(* Returns a counter-model of the formula
   in Godel-Dummett logic using a trace
   computation. The counter-model is not
   necessarily minimal.
*)

let compute_t ?(log=1) s =
  let _ = Compute.loglevel := log       in
  let f = Syntax.parse_string s     in
  let g = new Cgraph.form_graph f   in
  let _ = Cgraph.print g            
  in  Compute.trace g;;

(* Returns a counter-model of the formula
   in Godel-Dummett logic using a sum
   computation. The counter is minimal.
   Its size is the smallest integer
   n such that the formula is a non-
   theorem of LC(n).
*)

let compute_s ?(log=1) s =
  let _ = Compute.loglevel := log    in
  let f = Syntax.parse_string s     in
  let g = new Cgraph.form_graph f   in
  let _ = Cgraph.print g            
  in  Compute.sum g;;

(* Quite surprisingly, the sum computation
   seems to be more efficient than the
   trace computation, at least on these
   examples ... to be confirmed in the
   general case ...
*)

let chain i j =
  let a k = Printf.sprintf "a%d" k in
  let rec loop s i =
    if   i<j
    then loop (Printf.sprintf "%s|%s->%s" s (a i) (a (i+1))) (i+1)
    else s
  in loop (Printf.sprintf "%s->%s" (a i) (a (i+1))) (i+1);;

(* ******************* Examples ************************ *)

(* ff represents false and tt represents true *)

compute_t "((a->b)->a)->a";;
compute_s "((a->b)->a)->a";;

compute_t ~log:2 "((a->ff)->ff)->a";;
compute_s ~log:2 "((a->ff)->ff)->a";;

compute_t "a|a->ff";;
compute_s "a|a->ff";;

compute "(c|c->ff)&((a->b)->a)->a";;

compute_s "a->ff|(a->ff)->ff";;

compute "(a->b)|(b->a)";;

compute_t ~log:2 "(a1->a2)|(a2->a3)|(a3->a4)|(a4->a5)";;
compute_s ~log:2 "(a1->a2)|(a2->a3)|(a3->a4)|(a4->a5)";;

compute_t ~log:2 "(a1->a2)|(a2->a3)|(a3->a4)|(a4->a5)|(a5->a1)";;
compute_s ~log:2 "(a1->a2)|(a2->a3)|(a3->a4)|(a4->a5)|(a5->a1)";;

compute "(a1->a2)|(a2->a3)|(a3->a4)|(a4->a1)";;

compute_s  ~log:2  "(x|x|x|x)->(x&x&x&x)";;
compute_s  "(x|x|x|x|x|x|x|x|x|x|x)->(x&x&x&x&x&x&x&x&x&x&x&x&x&x&x&x&x)";;

compute_s "(x|x|x|x|x|x|y|x|x|x|x)->(x&x&x&x&x&x&x&x&x&x&x&x&x&x&x&x&x)";;

compute_s "(x&x&x&x&x&x&x&x&x&x&x&x&x&x&x&x&x)->(x|x|x|x|y|x|x|x|x|x|x)";;

compute_s ~log:2 "((a|b)&c)->((a&c)|(b&c))";;
compute_s "((a|b)&c)->((a&c)|(b&d))";;

compute_t (chain 1 15);;
compute_s (chain 1 15);;
compute_t ~log:2 ((chain 1 15) ^ "|a15->a1");;
compute_s ~log:2 ((chain 1 15) ^ "|a15->a1");;

compute_t (chain 1 30);;

compute_s ~log:2 (chain 1 30);;
compute_s (chain 1 100);;

compute_s ((chain 1 100) ^ "|a100->a1");;

compute_s "
       (e|g|h|k|l)
    -> (e|n|p|k|q)
    -> (e|g|k|l|q)
    -> (r|p|k|l|q)
    -> (s|r|p|l|q)
    -> (r|h|p|k|l)
    -> (r|n|g|h|k)
    -> (s|e|g|h|l)
    -> (e|n|h|p|k)
    -> (r|g|h|k|l)
    -> (e|p|k|l|q)
    -> (s|e|g|l|q)
    -> (s|r|h|p|l)
    -> (e|h|p|k|l)
    -> (e|n|g|h|k)
    -> (s|e|p|l|q)
    -> (s|r|n|g|h)
    -> (s|e|h|p|l)
    -> (s|e|n|g|q)
    -> (s|e|n|p|q)
    -> (r|g|k|l|q)
    -> (s|e|n|g|h)
    -> (r|n|p|k|q)
    -> (s|r|n|h|t)
    -> (r|n|g|k|q)
    -> (s|r|g|l|q)
    -> (s|e|n|h|t)
    -> (r|n|h|p|k)
    -> (s|r|g|h|l)
    -> (e|n|g|k|q)
    -> (s|r|n|g|q)
    -> (s|r|n|p|q)
    -> ((p&g)|(l&n)|(s&k)|(r&e)|(h&q)) ";;

compute_s "( 
      (e|g|h|k|l)
    & (e|n|p|k|q)
    & (e|g|k|l|q)
    & (r|p|k|l|q)
    & (s|r|p|l|q)
    & (r|h|p|k|l)
    & (r|n|g|h|k)
    & (s|e|g|h|l)
    & (e|n|h|p|k)
    & (r|g|h|k|l)
    & (e|p|k|l|q)
    & (s|e|g|l|q)
    & (s|r|h|p|l)
    & (e|h|p|k|l)
    & (e|n|g|h|k)
    & (s|e|p|l|q)
    & (s|r|n|g|h)
    & (s|e|h|p|l)
    & (s|e|n|g|q)
    & (s|e|n|p|q)
    & (r|g|k|l|q)
    & (s|e|n|g|h)
    & (r|n|p|k|q)
    & (s|r|n|h|t)
    & (r|n|g|k|q)
    & (s|r|g|l|q)
    & (s|e|n|h|t)
    & (r|n|h|p|k)
    & (s|r|g|h|l)
    & (e|n|g|k|q)
    & (s|r|n|g|q)
    & (s|r|n|p|q) )
    -> ((p&g)|(l&n)|(s&k)|(r&e)|(h&q)) ";;

compute "( 
      (p|l|s|r|h)
    & (p|l|s|r|q)
    & (p|l|s|e|h)
    & (p|l|s|e|q)
    & (p|l|k|r|h)
    & (p|l|k|r|q)
    & (p|l|k|e|h)
    & (p|l|k|e|q)
    & (p|n|s|r|h)
    & (p|n|s|r|q)
    & (p|n|s|e|h)
    & (p|n|s|e|q)
    & (p|n|k|r|h)
    & (p|n|k|r|q)
    & (p|n|k|e|h)
    & (p|n|k|e|q)
    & (g|l|s|r|h)
    & (g|l|s|r|q)
    & (g|l|s|e|h)
    & (g|l|s|e|q)
    & (g|l|k|r|h)
    & (g|l|k|r|q)
    & (g|l|k|e|h)
    & (g|l|k|e|q)
    & (g|n|s|r|h)
    & (g|n|s|r|q)
    & (g|n|s|e|h)
    & (g|n|s|e|q)
    & (g|n|k|r|h)
    & (g|n|k|r|q)
    & (g|n|k|e|h)
    & (g|n|k|e|q) )
    -> ((p&g)|(l&n)|(s&k)|(r&e)|(h&q)) ";;


compute "
       (a|b|c|d|e|f|g|h|i|j|k|l|m)
    -> (a|b|c|d|e|f|n|p|i|j|k|m|q)
    -> (a|b|c|d|e|f|g|i|j|k|l|m|q)
    -> (a|b|r|c|d|f|p|i|j|k|l|m|q)
    -> (a|s|b|r|c|d|f|p|i|j|l|m|q)
    -> (a|b|r|c|d|f|h|p|i|j|k|l|m)
    -> (a|b|r|c|d|f|n|g|h|i|j|k|m)
    -> (a|s|b|c|d|e|f|g|h|i|j|l|m)
    -> (a|b|c|d|e|f|n|h|p|i|j|k|m)
    -> (a|b|r|c|d|f|g|h|i|j|k|l|m)
    -> (a|b|c|d|e|f|p|i|j|k|l|m|q)
    -> (a|s|b|c|d|e|f|g|i|j|l|m|q)
    -> (a|s|b|r|c|d|f|h|p|i|j|l|m)
    -> (a|b|c|d|e|f|h|p|i|j|k|l|m)
    -> (a|b|c|d|e|f|n|g|h|i|j|k|m)
    -> (a|s|b|c|d|e|f|p|i|j|l|m|q)
    -> (a|s|b|r|c|d|f|n|g|h|i|j|m)
    -> (a|s|b|c|d|e|f|h|p|i|j|l|m)
    -> (a|s|b|c|d|e|f|n|g|i|j|m|q)
    -> (a|s|b|c|d|e|f|n|p|i|j|m|q)
    -> (a|b|r|c|d|f|g|i|j|k|l|m|q)
    -> (a|s|b|c|d|e|f|n|g|h|i|j|m)
    -> (a|b|r|c|d|f|n|p|i|j|k|m|q)
    -> (a|s|b|r|c|d|f|n|h|t|i|j|m)
    -> (a|b|r|c|d|f|n|g|i|j|k|m|q)
    -> (a|s|b|r|c|d|f|g|i|j|l|m|q)
    -> (a|s|b|c|d|e|f|n|h|t|i|j|m)
    -> (a|b|r|c|d|f|n|h|p|i|j|k|m)
    -> (a|s|b|r|c|d|f|g|h|i|j|l|m)
    -> (a|b|c|d|e|f|n|g|i|j|k|m|q)
    -> (a|s|b|r|c|d|f|n|g|i|j|m|q)
    -> (a|s|b|r|c|d|f|n|p|i|j|m|q)
    -> ((p&g)|(l&n)|(s&k)|(r&e)|(h&q)|a|b|c|d|f|i|j|m) ";;

compute_s "(
      (a|b|c|d|e|f|g|h|i|j|k|l|m)
    & (a|b|c|d|e|f|n|p|i|j|k|m|q)
    & (a|b|c|d|e|f|g|i|j|k|l|m|q)
    & (a|b|r|c|d|f|p|i|j|k|l|m|q)
    & (a|s|b|r|c|d|f|p|i|j|l|m|q)
    & (a|b|r|c|d|f|h|p|i|j|k|l|m)
    & (a|b|r|c|d|f|n|g|h|i|j|k|m)
    & (a|s|b|c|d|e|f|g|h|i|j|l|m)
    & (a|b|c|d|e|f|n|h|p|i|j|k|m)
    & (a|b|r|c|d|f|g|h|i|j|k|l|m)
    & (a|b|c|d|e|f|p|i|j|k|l|m|q)
    & (a|s|b|c|d|e|f|g|i|j|l|m|q)
    & (a|s|b|r|c|d|f|h|p|i|j|l|m)
    & (a|b|c|d|e|f|h|p|i|j|k|l|m)
    & (a|b|c|d|e|f|n|g|h|i|j|k|m)
    & (a|s|b|c|d|e|f|p|i|j|l|m|q)
    & (a|s|b|r|c|d|f|n|g|h|i|j|m)
    & (a|s|b|c|d|e|f|h|p|i|j|l|m)
    & (a|s|b|c|d|e|f|n|g|i|j|m|q)
    & (a|s|b|c|d|e|f|n|p|i|j|m|q)
    & (a|b|r|c|d|f|g|i|j|k|l|m|q)
    & (a|s|b|c|d|e|f|n|g|h|i|j|m)
    & (a|b|r|c|d|f|n|p|i|j|k|m|q)
    & (a|s|b|r|c|d|f|n|h|t|i|j|m)
    & (a|b|r|c|d|f|n|g|i|j|k|m|q)
    & (a|s|b|r|c|d|f|g|i|j|l|m|q)
    & (a|s|b|c|d|e|f|n|h|t|i|j|m)
    & (a|b|r|c|d|f|n|h|p|i|j|k|m)
    & (a|s|b|r|c|d|f|g|h|i|j|l|m)
    & (a|b|c|d|e|f|n|g|i|j|k|m|q)
    & (a|s|b|r|c|d|f|n|g|i|j|m|q)
    & (a|s|b|r|c|d|f|n|p|i|j|m|q) )
    -> ((p&g)|(l&n)|(s&k)|(r&e)|(h&q)|a|b|c|d|f|i|j|m) ";;

