open Bdd;;

type 'a column_t = 'a array;;
type 'a sparse_t = (int*'a) list;;
type 'a matrix_t = 'a sparse_t array;;

(* Compute the product between a sparse line l
   and an array column c and add s to the result.
*)

let scalprod_sl_c (b:bdd_t) l c s =
  let rec loop k = function
    | []        -> k
    | (i,v)::xs -> loop (b#add k v c.(i)) xs
  in loop s l;;

(* Compute the increasing product between a
   sparse row-matrix m and a column ci and
   put the result into column co.
   The result is the number of modified (ie
   strictly increased) cells.
   m is supposed to have top on every diagonal
   cell.
*)

let prodinc_sm_c b m ci co =
  let k     = ref 0 in 
  let f i l = 
    let xi  = ci.(i) in
    let yi  = scalprod_sl_c b l ci xi
    in  begin
          co.(i) <- yi;
          if   b#neq xi yi 
          then k := !k+1  
          else ()
        end 
  in begin
       Array.iteri f m; 
       !k
     end;;

(* Compute the product between a sparse 
   row-matrix m and a column ci and
   put the result into column co.
*)

let prod_sm_c b m ci co =
  let f i l = co.(i) <- scalprod_sl_c b l ci b#bot
  in  Array.iteri f m;;

let update_c f c = Array.iteri (fun i xi -> c.(i)<- f c.(i)) c;;

(* Add a constant k to each cell of a column c *)

let addconst_c b c k = update_c (b#disj k) c;;

(* Converts a a sparse column ci into an array column co *)

let sparseconv_sc_c b ci co =
  begin
    update_c  (fun _ -> b#bot) co;
    List.iter (fun (i,v) -> co.(i)<-v) ci
  end;;

(* Compute the empty and identity matrices *)

let matrix_empty (b:bdd_t) n = Array.create n [];;
let matrix_id (b:bdd_t) n    = let f i = [(i,b#top)] in Array.init n f;;

(* Computes the sum of to sparse rows *)

let sparsesum_r (b:bdd_t) k l =
  let rec loop r = function 
     | (k,[])              -> List.rev_append k r
     | ([],l)              -> List.rev_append l r
     | ((i,x)::k,(j,y)::l) -> if   i<j 
                              then loop ((i,x)::r) (k,(j,y)::l)
                              else if j<i
                              then loop ((j,y)::r) ((i,x)::k,l)
                              else loop ((i,b#disj x y)::r) (k,l)
  in List.rev (loop [] (k,l));;

(* Computes the sum of to sparse matrices *)

let sparsesum_m (b:bdd_t) m1 m2 =
  let n1   = Array.length m1 
  and n2   = Array.length m2 in
  let f1 i = if i<n1 then m1.(i) else []
  and f2 i = if i<n2 then m2.(i) else []
  and n    = max n1 n2 in
  let f i  = sparsesum_r b (f1 i) (f2 i)
  in  Array.init n f;;
