(**

  Memory allocation strategies

  A strategie takes the current size of a vector and returns the next size
  of the vector

  @author scott *)


(* hi *)

(** a strategy is a function that given the current size, returns the
  next size *)
type strategy = int -> int

(** create a constant strategy:  add constant i to size each time size increases *)
let const_strategy (i: int) =
  let strat sz = sz + i in strat

(** create a linear strategy: multiply size by constant i each time size increases *)
let linear_strategy (i: int) =
  let strat sz = sz * i in strat

(** create a log linear strategy: 
   take the log of the size, or 1, whichever is biggest, call it [d] and let the next
   size [n] be [c + ic/d] 
*)
let loglin_strategy (i: int) =
  let strat (sz: int) =
    let log_sz = log (float_of_int sz) in
    let coef_float = (float_of_int i) /. (max log_sz 1.0) in
      (int_of_float (ceil coef_float)) * sz + sz
  in
    strat

(** The default size sent to Array.create *)
let default_init_sz = 32

(** The default strategy *)
let default_strat = loglin_strategy 3

(** The type of a vector: each vector is an ocaml array and a strategy 
    and the total number of elements in it *)
type 'a t = { mutable a: 'a array; mutable total: int; init_elt: 'a; strat: strategy }

exception OutOfBounds of int * int

(** get and unsafe get *)
let get (vec: 'a t) (i: int) =  
  begin
    if i < vec.total then 
      Array.get vec.a i
    else
      raise (OutOfBounds(vec.total, Array.length vec.a))
  end

(** Unsafe_get can return the init_elt if the bound isn't less than the total *)
let unsafe_get (vec: 'a t) (i: int) =  
  Array.unsafe_get vec.a i


(** set and unsafe get 
  @param i the index of the vector to set 
  @param v the value to be set at the [i]th index 
  @return unit
*)
let set vec i v =
  begin
    if i < vec.total then 
      Array.set vec.a i v
    else
      raise (OutOfBounds(vec.total, Array.length vec.a))
  end


let unsafe_set (vec: 'a t) (i: int) (v: 'a) = 
  begin
    Array.unsafe_set vec.a i v
  end

(** create: supply an initializing value and a default size if you you*)
let create ?sz:(sz=default_init_sz) v =
  let ar = Array.create sz v in
    { a=ar; total=0; init_elt=v; strat=default_strat }

let realloc vec =
  let len_a = Array.length vec.a in
  let total_sz = vec.strat len_a in
  let new_ar_sz = total_sz - len_a in
  let new_ar = Array.create new_ar_sz vec.init_elt in
  vec.a <- Array.append vec.a new_ar

(** append 
@param vec the vector to which an element will be appended
@param elt the element to be appended.
@return unit return unit
*)
let append (vec: 'a t) elt =
  vec.total <- vec.total + 1;
  if vec.total > Array.length vec.a then
    realloc vec;
  Array.set vec.a (vec.total - 1) elt

(** 
  concatentate two vectors 
  @param vec1 the first vector
  @param vec2 the second vector
  @return bdd return a new vector consisting of the elements of vec1 followed by those of vec2 
*)
let concat (vec1: 'a t) (vec2: 'a t) =
  let newarry = Array.append vec1.a vec2.a in
    Array.blit newarry (Array.length vec1.a) newarry vec1.total vec2.total;
    { 
      a=newarry; 
      total=vec1.total + vec2.total; 
      strat=vec1.strat; 
      init_elt=vec1.init_elt;
    }
    
(** create a sub vector 
@param vec the vector from which to take a sub vector
@param start the start index of the vector
@param len  the length of the sub vector ([start + len] is the end index )
@return vector returns a vector object
*)
let sub (vec: 'a t) start len =
  let newary = Array.sub vec.a start len in
    { 
      a=newary;
      total=len;
      strat=vec.strat;
      init_elt=vec.init_elt
    }

(** copy a vector 
@param vec the vector to be copied
@return vec a copy of the argument
*)
let copy (vec: 'a t) =
  let newary = Array.copy vec.a in
    {
      a = newary;
      total=vec.total;
      strat=vec.strat;
      init_elt=vec.init_elt
    }

(** return the lenght of a vector *)
let length (vec: 'a t) =  vec.total

(** fill a vector with a value from [start] for [length] elts 
  @param vec the vector to be filled
  @param start the start index from which to fill
  @param len  fill for len elements
  @return unit operate by side affect
*)
(* XXX don't use while, it's slower than it needs to be *)
let fill (vec: 'a t) (start: int) (len: int) (elt: 'a) =
  while (start + len < Array.length vec.a) do realloc vec; done;
  Array.fill vec.a start len elt;
  if (start + len > vec.total) then
    vec.total <- start + len

let blit (vec1: 'a t) (o1: int) (vec2: 'a t) (o2: int) (len: int) =
  if ((o1 + len) >= vec1.total || (o2 + len) >= vec2.total) then
    raise (OutOfBounds(vec1.total, Array.length vec1.a))
  Array.blit vec1.a o1 vec2.a o2 len

(** turn a vector into a list *)
let to_list (vec: 'a t) =  
  Array.to_list (Array.sub vec.a 0 vec.total)

(** create a vector from a list *)
let of_list (lst: 'a list) =
  if List.length lst = 0 then
    raise (OutOfBounds(0,0));
  let ary = Array.of_list lst in
    { 
      a=ary;
      total=Array.length ary;
      strat=default_strat;
      init_elt=ary.(0);
    }

(** iterate a function over the elements of the list
  @param f the function to be iterated
  @param vec the vector to be iterated over
  @return () operate by side effect *)
let iter f vec =  Array.iter f (Array.sub vec.a 0 vec.total)

(** iterate a function over the elements and indices of the list
  @param f the function to be iterated
  @param vec the vector to be iterated over
  @return () operate by side effect *)
let iteri f vec = Array.iteri f (Array.sub vec.a 0 vec.total)

(** map a function over a vector
  @param f the function to be mapped
  @param vec the vector whose elements are the argument of the function
  @return vec return a vector whose values are f(x) for each x in vec *)
let map f vec =
  let new_ary = Array.map f (Array.sub vec.a 0 vec.total) in
  let new_init_elt = f vec.init_elt in
    {
      a=new_ary;
      total=vec.total;
      strat=vec.strat;
      init_elt=new_init_elt;
    }

let mapi f vec =
  let new_ary = Array.mapi f (Array.sub vec.a 0 vec.total) in
  let new_init_elt = f 0 vec.init_elt in
    {
      a=new_ary;
      total=vec.total;
      strat=vec.strat;
      init_elt=new_init_elt;
    }

let sort f vec =  Array.sort f vec.a
let stable_sort f vec = Array.stable_sort f vec.a
let fast_sort f vec = Array.fast_sort f vec.a




  





  
